/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/**
 *  \file stat.c Functions used for sending/receiving the status of para_server
 *  and para_audiod.
 */

#include <regex.h>

#include "para.h"
#include "error.h"
#include "string.h"


/** The minimal length of a status item buffer. */
#define MIN_STAT_ITEM_LEN 9 /* 5 + 2 + 2, e.g. '0005 00:\n' */

/**
 * Call a function for each complete status item of a buffer.
 *
 * \param item_buf The source buffer.
 * \param num_bytes The length of \a buf.
 * \param item_handler Function to call for each complete item.
 *
 * \return Negative on errors, the number of bytes _not_ passed to \a
 * item_handler on success.
 *
 * Status items are expected in the format used by parser-friendly output mode
 * of the stat command of para_client/para_audioc.
 */
int for_each_stat_item(char *item_buf, size_t num_bytes,
	int (*item_handler)(int, char *))
{
	char *buf = item_buf;
	int len = num_bytes;

	for (;;) {
		int i, ret, item_len, item_num = 0;
		if (len < MIN_STAT_ITEM_LEN)
			break;
		ret = read_size_header(buf);
		if (ret < 0)
			return ret;
		item_len = ret;
		if (item_len > len - 5) /* item not complete */
			break;
		for (i = 0; i < 2; i++) {
			unsigned char c = buf[5 + i];
			item_num <<= 4;
			if (c >= '0' && c <= '9') {
				item_num += c - '0';
				continue;
			}
			if (c >= 'a' && c <= 'f') {
				item_num += c - 'a' + 10;
				continue;
			}
			return -E_STAT_ITEM_PARSE;
		}
		if (buf[7] != ':' || buf[5 + item_len - 1] != '\n')
			return -E_STAT_ITEM_PARSE;
		buf[5 + item_len - 1] = '\0';
		if (item_num >= NUM_STAT_ITEMS)
			PARA_WARNING_LOG("unknown status item %d: %s\n",
				item_num, buf + 8);
		else {
			ret = item_handler(item_num, buf + 8);
			if (ret < 0)
				return ret;
		}
		buf += 5 + item_len;
		len -= 5 + item_len;
		assert(len >= 0 && buf <= item_buf + num_bytes);
	}
	assert(len >= 0);
	return len;
}
