/*
 * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file stdin.c Functions that deal with reading from stdin. */

#include <regex.h>

#include "para.h"
#include "list.h"
#include "sched.h"
#include "fd.h"
#include "error.h"
#include "stdin.h"
#include "buffer_tree.h"
#include "string.h"

/*
 * If there is space left in the buffer of the stdin task add STDIN_FILENO to
 * the read fd set of s.
 */
static void stdin_pre_select(struct sched *s, void *context)
{
	struct stdin_task *sit = context;
	int ret;

	ret = btr_node_status(sit->btrn, 0, BTR_NT_ROOT);
	if (ret < 0)
		sched_min_delay(s);
	if (ret <= 0)
		return;
	if (btr_pool_unused(sit->btrp) > 0)
		return para_fd_set(STDIN_FILENO, &s->rfds, &s->max_fileno);
	sched_request_timeout_ms(100, s);
}

/*
 * This function checks if STDIN_FILENO was included by in the read fd set of s
 * during the previous pre_select call. If so, and if STDIN_FILENO is readable,
 * data is read from stdin and fed into the buffer tree.
 */
static int stdin_post_select(struct sched *s, void *context)
{
	struct stdin_task *sit = context;
	ssize_t ret;
	size_t sz, n;
	char *buf = NULL;

	ret = btr_node_status(sit->btrn, 0, BTR_NT_ROOT);
	if (ret < 0)
		goto err;
	if (ret == 0)
		return 0;
	sz = btr_pool_get_buffer(sit->btrp, &buf);
	if (sz == 0)
		return 0;
	if (sit->must_set_nonblock_flag) {
		ret = mark_fd_nonblocking(STDIN_FILENO);
		if (ret < 0)
			goto err;
		sit->must_set_nonblock_flag = false;
	}
	/*
	 * Do not use the maximal size to avoid having only a single buffer
	 * reference for the whole pool. This is bad because if that single
	 * reference can not be freed, we're stuck.
	 */
	sz = PARA_MIN(sz, btr_pool_size(sit->btrp) / 2);
	ret = read_nonblock(STDIN_FILENO, buf, sz, &s->rfds, &n);
	if (n > 0)
		btr_add_output_pool(sit->btrp, n, sit->btrn);
	if (ret >= 0)
		return 0;
err:
	btr_remove_node(&sit->btrn);
	/* Revert to blocking mode if necessary. */
	fcntl(STDIN_FILENO, F_SETFL, sit->fd_flags);
	//btr_pool_free(sit->btrp);
	return ret;
}

/**
 * Register a stdin task structure.
 *
 * \param sit The stdin task structure to register.
 * \param s The task will be added to this scheduler's task list.
 *
 * This allocates a buffer tree pool for I/O, sets up \a sit and registers a
 * task with \a sit as context pointer.
 */
void stdin_task_register(struct stdin_task *sit, struct sched *s)
{
	int ret;
	struct task_info ti = {
		.name = "stdin",
		.pre_select = stdin_pre_select,
		.post_select = stdin_post_select,
		.context = sit,
	};

	sit->btrp = btr_pool_new("stdin", 128 * 1024);
	/*
	 * Both STDIN_FILENO and STDOUT_FILENO may refer to the same open file
	 * description (the terminal), and thus share the same file status
	 * flags. In order to not interfere with the stdout task, we only get
	 * the file status flags for STDIN here and save a copy. The nonblock
	 * flag is set later on the first read.
	 */
	ret = fcntl(STDIN_FILENO, F_GETFL);
	if (ret < 0) {
		PARA_EMERG_LOG("F_GETFL: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	sit->fd_flags = ret;
	sit->must_set_nonblock_flag = (sit->fd_flags & O_NONBLOCK) == 0;
	sit->task = task_register(&ti, s);
}
